import pytest
from pytest import approx
from math import radians, inf

import pymap3d.latitude as latitude
import pymap3d.rcurve as rcurve


@pytest.mark.parametrize(
    "geodetic_lat,alt_m,geocentric_lat",
    [(0, 0, 0), (90, 0, 90), (-90, 0, -90), (45, 0, 44.80757678), (-45, 0, -44.80757678)],
)
def test_geodetic_alt_geocentric(geodetic_lat, alt_m, geocentric_lat):
    assert latitude.geod2geoc(geodetic_lat, alt_m) == approx(geocentric_lat)

    r = rcurve.geocentric_radius(geodetic_lat)
    assert latitude.geoc2geod(geocentric_lat, r) == approx(geodetic_lat)
    assert latitude.geoc2geod(geocentric_lat, 1e5 + r) == approx(
        latitude.geocentric2geodetic(geocentric_lat, 1e5 + alt_m)
    )

    assert latitude.geod2geoc(geodetic_lat, 1e5 + alt_m) == approx(
        latitude.geodetic2geocentric(geodetic_lat, 1e5 + alt_m)
    )


@pytest.mark.parametrize(
    "geodetic_lat,geocentric_lat",
    [(0, 0), (90, 90), (-90, -90), (45, 44.80757678), (-45, -44.80757678)],
)
def test_geodetic_geocentric(geodetic_lat, geocentric_lat):

    assert latitude.geodetic2geocentric(geodetic_lat, 0) == approx(geocentric_lat)
    assert latitude.geodetic2geocentric(radians(geodetic_lat), 0, deg=False) == approx(
        radians(geocentric_lat)
    )

    assert latitude.geocentric2geodetic(geocentric_lat, 0) == approx(geodetic_lat)
    assert latitude.geocentric2geodetic(radians(geocentric_lat), 0, deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_geocentric():
    pytest.importorskip("numpy")
    assert latitude.geodetic2geocentric([45, 0], 0) == approx([44.80757678, 0])
    assert latitude.geocentric2geodetic([44.80757678, 0], 0) == approx([45, 0])


@pytest.mark.parametrize(
    "geodetic_lat, isometric_lat",
    [(0, 0), (90, inf), (-90, -inf), (45, 50.227466), (-45, -50.227466), (89, 271.275)],
)
def test_geodetic_isometric(geodetic_lat, isometric_lat):
    isolat = latitude.geodetic2isometric(geodetic_lat)
    assert isolat == approx(isometric_lat)
    assert isinstance(isolat, float)

    assert latitude.geodetic2isometric(radians(geodetic_lat), deg=False) == approx(
        radians(isometric_lat)
    )

    assert latitude.isometric2geodetic(isometric_lat) == approx(geodetic_lat)
    assert latitude.isometric2geodetic(radians(isometric_lat), deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_isometric():
    pytest.importorskip("numpy")
    assert latitude.geodetic2isometric([45, 0]) == approx([50.227466, 0])
    assert latitude.isometric2geodetic([50.227466, 0]) == approx([45, 0])


@pytest.mark.parametrize(
    "geodetic_lat,conformal_lat",
    [(0, 0), (90, 90), (-90, -90), (45, 44.80768406), (-45, -44.80768406), (89, 88.99327)],
)
def test_geodetic_conformal(geodetic_lat, conformal_lat):
    clat = latitude.geodetic2conformal(geodetic_lat)
    assert clat == approx(conformal_lat)
    assert isinstance(clat, float)

    assert latitude.geodetic2conformal(radians(geodetic_lat), deg=False) == approx(
        radians(conformal_lat)
    )

    assert latitude.conformal2geodetic(conformal_lat) == approx(geodetic_lat)
    assert latitude.conformal2geodetic(radians(conformal_lat), deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_conformal():
    pytest.importorskip("numpy")
    assert latitude.geodetic2conformal([45, 0]) == approx([44.80768406, 0])
    assert latitude.conformal2geodetic([44.80768406, 0]) == approx([45, 0])


@pytest.mark.parametrize(
    "geodetic_lat,rectifying_lat",
    [(0, 0), (90, 90), (-90, -90), (45, 44.855682), (-45, -44.855682)],
)
def test_geodetic_rectifying(geodetic_lat, rectifying_lat):
    assert latitude.geodetic2rectifying(geodetic_lat) == approx(rectifying_lat)
    assert latitude.geodetic2rectifying(radians(geodetic_lat), deg=False) == approx(
        radians(rectifying_lat)
    )

    assert latitude.rectifying2geodetic(rectifying_lat) == approx(geodetic_lat)
    assert latitude.rectifying2geodetic(radians(rectifying_lat), deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_rectifying():
    pytest.importorskip("numpy")
    assert latitude.geodetic2rectifying([45, 0]) == approx([44.855682, 0])
    assert latitude.rectifying2geodetic([44.855682, 0]) == approx([45, 0])


@pytest.mark.parametrize(
    "geodetic_lat,authalic_lat",
    [(0, 0), (90, 90), (-90, -90), (45, 44.87170288), (-45, -44.87170288)],
)
def test_geodetic_authalic(geodetic_lat, authalic_lat):
    assert latitude.geodetic2authalic(geodetic_lat) == approx(authalic_lat)
    assert latitude.geodetic2authalic(radians(geodetic_lat), deg=False) == approx(
        radians(authalic_lat)
    )

    assert latitude.authalic2geodetic(authalic_lat) == approx(geodetic_lat)
    assert latitude.authalic2geodetic(radians(authalic_lat), deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_authalic():
    pytest.importorskip("numpy")
    assert latitude.geodetic2authalic([45, 0]) == approx([44.87170288, 0])
    assert latitude.authalic2geodetic([44.87170288, 0]) == approx([45, 0])


@pytest.mark.parametrize(
    "geodetic_lat,parametric_lat",
    [(0, 0), (90, 90), (-90, -90), (45, 44.9037878), (-45, -44.9037878)],
)
def test_geodetic_parametric(geodetic_lat, parametric_lat):
    assert latitude.geodetic2parametric(geodetic_lat) == approx(parametric_lat)
    assert latitude.geodetic2parametric(radians(geodetic_lat), deg=False) == approx(
        radians(parametric_lat)
    )

    assert latitude.parametric2geodetic(parametric_lat) == approx(geodetic_lat)
    assert latitude.parametric2geodetic(radians(parametric_lat), deg=False) == approx(
        radians(geodetic_lat)
    )


def test_numpy_geodetic_parametric():
    pytest.importorskip("numpy")
    assert latitude.geodetic2parametric([45, 0]) == approx([44.9037878, 0])
    assert latitude.parametric2geodetic([44.9037878, 0]) == approx([45, 0])


@pytest.mark.parametrize("lat", [91, -91])
def test_badvals(lat):
    # geodetic_isometric is not included on purpose
    with pytest.raises(ValueError):
        latitude.geodetic2geocentric(lat, 0)
    with pytest.raises(ValueError):
        latitude.geocentric2geodetic(lat, 0)
    with pytest.raises(ValueError):
        latitude.geodetic2conformal(lat)
    with pytest.raises(ValueError):
        latitude.conformal2geodetic(lat)
    with pytest.raises(ValueError):
        latitude.geodetic2rectifying(lat)
    with pytest.raises(ValueError):
        latitude.rectifying2geodetic(lat)
    with pytest.raises(ValueError):
        latitude.geodetic2authalic(lat)
    with pytest.raises(ValueError):
        latitude.authalic2geodetic(lat)
    with pytest.raises(ValueError):
        latitude.geodetic2parametric(lat)
    with pytest.raises(ValueError):
        latitude.parametric2geodetic(lat)



from .. import vincenty
import argparse


p = argparse.ArgumentParser(
    description="Given starting latitude, longitude: find final lat,lon for distance and azimuth"
)
p.add_argument("lat", help="latitude WGS-84 [degrees]", type=float)
p.add_argument("lon", help="longitude WGS-84 [degrees]", type=float)
p.add_argument("range", help="range from start point [meters]", type=float)
p.add_argument("azimuth", help="clockwise from north: azimuth to start [degrees]", type=float)
P = p.parse_args()

lat2, lon2 = vincenty.vreckon(P.lat, P.lon, P.range, P.azimuth)

print("lat, lon = ({:.4f}, {:.4f})".format(lat2.item(), lon2.item()))



import pytest
from pytest import approx

import pymap3d.lox as lox


@pytest.mark.parametrize("lat,dist", [(0, 0), (90, 10001965.729)])
def test_meridian_dist(lat, dist):
    assert lox.meridian_dist(lat) == approx(dist)


@pytest.mark.parametrize(
    "lat1,lat2,arclen",
    [
        (0, 0, 0),
        (0, 90, 10001965.729),
        (0, -90, 10001965.729),
        (0, 40, 4429529.03035058),
        (40, 80, 4455610.84159),
    ],
)
def test_meridian_arc(lat1, lat2, arclen):
    """
    meridianarc(deg2rad(40), deg2rad(80), wgs84Ellipsoid)
    """

    assert lox.meridian_arc(lat1, lat2) == approx(arclen)


@pytest.mark.parametrize(
    "lon1,lon2,lat,dist",
    [
        (0, 0, 0, 0),
        (0, 90, 0, 10018754.1714),
        (0, -90, 0, 10018754.1714),
        (90, 0, 0, 10018754.1714),
        (-90, 0, 0, 10018754.1714),
    ],
)
def test_departure(lon1, lon2, lat, dist):
    assert lox.departure(lon1, lon2, lat) == approx(dist)


@pytest.mark.parametrize(
    "lat1,lon1,lat2,lon2,arclen,az",
    [
        (40, -80, 65, -148, 5248666.20853187, 302.0056736),
        (0, 0, 0, 90, 10018754.17, 90),
        (0, 0, 0, -90, 10018754.17, 270),
        (0, 90, 0, 0, 10018754.17, 270),
        (0, -90, 0, 0, 10018754.17, 90),
        (1, 0, 0, 0, 110574.4, 180),
        (-1, 0, 0, 0, 110574.4, 0),
    ],
)
def test_loxodrome_inverse(lat1, lon1, lat2, lon2, arclen, az):
    """
    distance('rh', 40, -80, 65, -148, wgs84Ellipsoid)
    azimuth('rh', 40, -80, 65, -148, wgs84Ellipsoid)
    """
    rhdist, rhaz = lox.loxodrome_inverse(lat1, lon1, lat2, lon2)

    assert rhdist == approx(arclen)
    assert rhaz == approx(az)
    assert isinstance(rhdist, float)
    assert isinstance(rhaz, float)


def test_numpy_loxodrome_inverse():
    pytest.importorskip("numpy")
    d, a = lox.loxodrome_inverse([40, 40], [-80, -80], 65, -148)

    assert d == approx(5248666.209)
    assert a == approx(302.00567)


@pytest.mark.parametrize(
    "lat0,lon0,rng,az,lat1,lon1",
    [
        (40, -80, 10000, 30, 40.077995, -79.9414144),
        (0, 0, 0, 0, 0, 0),
        (0, 0, 10018754.17, 90, 0, 90),
        (0, 0, 10018754.17, -90, 0, -90),
        (0, 0, 110574.4, 180, -1, 0),
        (-1, 0, 110574.4, 0, 0, 0),
    ],
)
def test_loxodrome_direct(lat0, lon0, rng, az, lat1, lon1):
    lat2, lon2 = lox.loxodrome_direct(lat0, lon0, rng, az)
    assert lat2 == approx(lat1, abs=1e-6)
    assert lon2 == approx(lon1)
    assert isinstance(lat2, float)
    assert isinstance(lon2, float)


def test_numpy_loxodrome_direct():
    pytest.importorskip("numpy")
    lat, lon = lox.loxodrome_direct([40, 40], [-80, -80], [10000, 10000], [30, 30])
    assert lat == approx(40.077995)
    assert lon == approx(-79.941414)


@pytest.mark.parametrize("lat,lon", [([0, 45, 90], [0, 45, 90])])
def test_meanm(lat, lon):
    pytest.importorskip("numpy")
    assert lox.meanm(lat, lon) == approx([47.26967, 18.460557])



from math import radians
import pytest
from pytest import approx

import pymap3d as pm

ELL = pm.Ellipsoid()
A = ELL.semimajor_axis
B = ELL.semiminor_axis


@pytest.mark.parametrize("xyz", [(0, A, 50), ([0], [A], [50])], ids=("scalar", "list"))
def test_scalar_enu(xyz):
    """
    verify we can handle the wide variety of input data type users might use
    """
    if isinstance(xyz[0], list):
        pytest.importorskip("numpy")

    enu = pm.ecef2enu(*xyz, 0, 90, -100)

    assert pm.enu2ecef(*enu, 0, 90, -100) == approx([0, A, 50])


def test_3d_enu():
    np = pytest.importorskip("numpy")
    xyz = (np.atleast_3d(0), np.atleast_3d(A), np.atleast_3d(50))

    enu = pm.ecef2enu(*xyz, 0, 90, -100)
    assert pm.enu2ecef(*enu, 0, 90, -100) == approx([0, A, 50])


@pytest.mark.parametrize(
    "enu,lla,xyz", [((0, 0, 0), (0, 0, 0), (A, 0, 0)), ((0, 0, 1000), (0, 0, 0), (A + 1000, 0, 0))]
)
def test_enu_ecef(enu, lla, xyz):
    x, y, z = pm.enu2ecef(*enu, *lla)
    assert x == approx(xyz[0])
    assert y == approx(xyz[1])
    assert z == approx(xyz[2])
    assert isinstance(x, float)
    assert isinstance(y, float)
    assert isinstance(z, float)

    rlla = (radians(lla[0]), radians(lla[1]), lla[2])
    assert pm.enu2ecef(*enu, *rlla, deg=False) == approx(xyz)

    e, n, u = pm.ecef2enu(*xyz, *lla)
    assert e == approx(enu[0])
    assert n == approx(enu[1])
    assert u == approx(enu[2])
    assert isinstance(e, float)
    assert isinstance(n, float)
    assert isinstance(u, float)

    e, n, u = pm.ecef2enu(*xyz, *rlla, deg=False)
    assert e == approx(enu[0])
    assert n == approx(enu[1])
    assert u == approx(enu[2])



"""Minimal class for planetary ellipsoids"""
from math import sqrt


class Ellipsoid:
    """
    generate reference ellipsoid parameters

    https://en.wikibooks.org/wiki/PROJ.4#Spheroid

    https://nssdc.gsfc.nasa.gov/planetary/factsheet/index.html

    as everywhere else in this program, distance units are METERS
    """

    def __init__(self, model: str = "wgs84"):
        """
        feel free to suggest additional ellipsoids

        Parameters
        ----------
        model : str
                name of ellipsoid
        """
        if model == "wgs84":
            """https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84"""
            self.semimajor_axis = 6378137.0
            self.semiminor_axis = 6356752.31424518
        elif model == "wgs72":
            self.semimajor_axis = 6378135.0
            self.semiminor_axis = 6356750.52001609
        elif model == "grs80":
            """https://en.wikipedia.org/wiki/GRS_80"""
            self.semimajor_axis = 6378137.0
            self.semiminor_axis = 6356752.31414036
        elif model == "clarke1866":
            self.semimajor_axis = 6378206.4
            self.semiminor_axis = 6356583.8
        elif model == "mars":
            """
            https://tharsis.gsfc.nasa.gov/geodesy.html
            """
            self.semimajor_axis = 3396900.0
            self.semiminor_axis = 3376097.80585952
        elif model == "moon":
            self.semimajor_axis = 1738000.0
            self.semiminor_axis = self.semimajor_axis
        elif model == "venus":
            self.semimajor_axis = 6051000.0
            self.semiminor_axis = self.semimajor_axis
        elif model == "jupiter":
            self.semimajor_axis = 71492000.0
            self.semiminor_axis = 66770054.3475922
        elif model == "io":
            """
            https://doi.org/10.1006/icar.1998.5987
            """
            self.semimajor_axis = 1829.7
            self.semiminor_axis = 1815.8
        elif model == "pluto":
            self.semimajor_axis = 1187000.0
            self.semiminor_axis = self.semimajor_axis
        else:
            raise NotImplementedError(
                "{} model not implemented, let us know and we will add it (or make a pull request)".format(
                    model
                )
            )

        self.flattening = (self.semimajor_axis - self.semiminor_axis) / self.semimajor_axis
        self.thirdflattening = (self.semimajor_axis - self.semiminor_axis) / (
            self.semimajor_axis + self.semiminor_axis
        )
        self.eccentricity = sqrt(2 * self.flattening - self.flattening ** 2)



import pytest
from pytest import approx
from math import radians, nan, sqrt, isnan

import pymap3d as pm

lla0 = (42, -82, 200)
rlla0 = (radians(lla0[0]), radians(lla0[1]), lla0[2])

xyz0 = (660675.2518247, -4700948.68316, 4245737.66222)

ELL = pm.Ellipsoid()
A = ELL.semimajor_axis
B = ELL.semiminor_axis

atol_dist = 1e-6  # 1 micrometer


@pytest.mark.parametrize("lla", [(42, -82, 200), ([42], [-82], [200])], ids=("scalar", "list"))
def test_scalar_geodetic2ecef(lla):
    """
    verify we can handle the wide variety of input data type users might use
    """
    if isinstance(lla[0], list):
        pytest.importorskip("numpy")

    x0, y0, z0 = pm.geodetic2ecef(*lla)

    assert (x0, y0, z0) == approx(xyz0)


def test_3d_geodetic2ecef():
    np = pytest.importorskip("numpy")
    lla = (np.atleast_3d(42), np.atleast_3d(-82), np.atleast_3d(200))
    x0, y0, z0 = pm.geodetic2ecef(*lla)

    assert (x0, y0, z0) == approx(xyz0)


def test_scalar_ecef2geodetic():
    """
    verify we can handle the wide variety of input data type users might use
    """
    lat, lon, alt = pm.ecef2geodetic(xyz0[0], xyz0[1], xyz0[2])

    assert [lat, lon, alt] == approx(lla0, rel=1e-4)


def test_3d_ecef2geodetic():
    np = pytest.importorskip("numpy")
    xyz = (np.atleast_3d(xyz0[0]), np.atleast_3d(xyz0[1]), np.atleast_3d(xyz0[2]))

    lat, lon, alt = pm.ecef2geodetic(*xyz)

    assert [lat, lon, alt] == approx(lla0, rel=1e-4)


def test_array_ecef2geodetic():
    """
    tests ecef2geodetic can handle numpy array data in addition to singular floats
    """
    np = pytest.importorskip("numpy")
    # test values with no points inside ellipsoid
    lla0_array = (
        np.array([lla0[0], lla0[0]]),
        np.array([lla0[1], lla0[1]]),
        np.array([lla0[2], lla0[2]]),
    )
    xyz = pm.geodetic2ecef(*lla0_array)
    lats, lons, alts = pm.ecef2geodetic(*xyz)

    assert lats == approx(lla0_array[0])
    assert lons == approx(lla0_array[1])
    assert alts == approx(lla0_array[2])

    # test values with some (but not all) points inside ellipsoid
    lla0_array_inside = (
        np.array([lla0[0], lla0[0]]),
        np.array([lla0[1], lla0[1]]),
        np.array([lla0[2], -lla0[2]]),
    )
    xyz = pm.geodetic2ecef(*lla0_array_inside)
    lats, lons, alts = pm.ecef2geodetic(*xyz)

    assert lats == approx(lla0_array_inside[0])
    assert lons == approx(lla0_array_inside[1])
    assert alts == approx(lla0_array_inside[2])


def test_xarray():
    xarray = pytest.importorskip("xarray")
    xr_lla = xarray.DataArray(list(lla0))

    xyz = pm.geodetic2ecef(*xr_lla)

    assert xyz == approx(xyz0)
    # %%
    xr_xyz = xarray.DataArray(list(xyz0))

    lla = pm.ecef2geodetic(*xr_xyz)

    assert lla == approx(lla0)


def test_pandas():
    pandas = pytest.importorskip("pandas")
    pd_lla = pandas.Series(lla0)

    xyz = pm.geodetic2ecef(*pd_lla)

    assert xyz == approx(xyz0)
    # %% dataframe degenerates to series
    pd_lla = pandas.DataFrame([[*lla0], [*lla0]], columns=["lat", "lon", "alt_m"])
    xyz = pm.geodetic2ecef(pd_lla["lat"], pd_lla["lon"], pd_lla["alt_m"])

    assert xyz[0].values == approx(xyz0[0])
    assert xyz[1].values == approx(xyz0[1])
    assert xyz[2].values == approx(xyz0[2])


def test_ecef():
    xyz = pm.geodetic2ecef(*lla0)

    assert xyz == approx(xyz0)
    x, y, z = pm.geodetic2ecef(*rlla0, deg=False)
    assert x == approx(xyz[0])
    assert y == approx(xyz[1])
    assert z == approx(xyz[2])

    with pytest.raises(ValueError):
        pm.geodetic2ecef(-100, lla0[1], lla0[2])

    assert pm.ecef2geodetic(*xyz) == approx(lla0)
    assert pm.ecef2geodetic(*xyz, deg=False) == approx(rlla0)

    assert pm.ecef2geodetic((A - 1) / sqrt(2), (A - 1) / sqrt(2), 0) == approx([0, 45, -1])


@pytest.mark.parametrize(
    "lla, xyz",
    [
        ((0, 0, -1), (A - 1, 0, 0)),
        ((0, 90, -1), (0, A - 1, 0)),
        ((0, -90, -1), (0, -A + 1, 0)),
        ((90, 0, -1), (0, 0, B - 1)),
        ((90, 15, -1), (0, 0, B - 1)),
        ((-90, 0, -1), (0, 0, -B + 1)),
    ],
)
def test_geodetic2ecef(lla, xyz):
    assert pm.geodetic2ecef(*lla) == approx(xyz, abs=atol_dist)


@pytest.mark.parametrize(
    "xyz, lla",
    [
        ((A - 1, 0, 0), (0, 0, -1)),
        ((0, A - 1, 0), (0, 90, -1)),
        ((0, 0, B - 1), (90, 0, -1)),
        ((0, 0, -B + 1), (-90, 0, -1)),
        ((-A + 1, 0, 0), (0, 180, -1)),
    ],
)
def test_ecef2geodetic(xyz, lla):
    lat, lon, alt = pm.ecef2geodetic(*xyz)
    assert lat == approx(lla[0])
    assert lon == approx(lla[1])
    assert alt == approx(lla[2])


@pytest.mark.parametrize(
    "aer,lla,lla0",
    [
        ((33, 77, 1000), (42.0016981935, -81.99852, 1174.374035), (42, -82, 200)),
        ((0, 90, 10000), (0, 0, 10000), (0, 0, 0)),
    ],
)
def test_aer_geodetic(aer, lla, lla0):
    lat1, lon1, alt1 = pm.aer2geodetic(*aer, *lla0)
    assert lat1 == approx(lla[0])
    assert lon1 == approx(lla[1])
    assert alt1 == approx(lla[2])
    assert isinstance(lat1, float)
    assert isinstance(lon1, float)
    assert isinstance(alt1, float)

    raer = (radians(aer[0]), radians(aer[1]), aer[2])
    rlla0 = (radians(lla0[0]), radians(lla0[1]), lla0[2])
    assert pm.aer2geodetic(*raer, *rlla0, deg=False) == approx(
        (radians(lla[0]), radians(lla[1]), lla[2])
    )

    with pytest.raises(ValueError):
        pm.aer2geodetic(aer[0], aer[1], -1, *lla0)

    assert pm.geodetic2aer(*lla, *lla0) == approx(aer, rel=1e-3)
    assert pm.geodetic2aer(radians(lla[0]), radians(lla[1]), lla[2], *rlla0, deg=False) == approx(
        raer, rel=1e-3
    )


def test_scalar_nan():
    a, e, r = pm.geodetic2aer(nan, nan, nan, *lla0)
    assert isnan(a) and isnan(e) and isnan(r)

    lat, lon, alt = pm.aer2geodetic(nan, nan, nan, *lla0)
    assert isnan(lat) and isnan(lon) and isnan(alt)


def test_allnan():
    np = pytest.importorskip("numpy")
    anan = np.empty((10, 10))
    anan.fill(nan)
    assert np.isnan(pm.geodetic2aer(anan, anan, anan, *lla0)).all()
    assert np.isnan(pm.aer2geodetic(anan, anan, anan, *lla0)).all()


def test_somenan():
    np = pytest.importorskip("numpy")
    xyz = np.stack((xyz0, (nan, nan, nan)))

    lat, lon, alt = pm.ecef2geodetic(xyz[:, 0], xyz[:, 1], xyz[:, 2])
    assert (lat[0], lon[0], alt[0]) == approx(lla0)



"""Utility functions

all assume radians"""

from __future__ import annotations
import typing

from .ellipsoid import Ellipsoid

try:
    from numpy import hypot, cos, sin, arctan2 as atan2, radians, pi, asarray, ndarray, sign
except ImportError:
    from math import atan2, hypot, cos, sin, radians, pi  # type: ignore

    asarray = None
    ndarray = typing.Any  # type: ignore

    def sign(x: float) -> float:  # type: ignore
        """signum function"""
        if x < 0:
            y = -1.0
        elif x > 0:
            y = 1.0
        else:
            y = 0.0

        return y


__all__ = ["cart2pol", "pol2cart", "cart2sph", "sph2cart", "sign"]


def cart2pol(x: float, y: float) -> tuple[float, float]:
    """Transform Cartesian to polar coordinates"""
    return atan2(y, x), hypot(x, y)


def pol2cart(theta: float, rho: float) -> tuple[float, float]:
    """Transform polar to Cartesian coordinates"""
    return rho * cos(theta), rho * sin(theta)


def cart2sph(
    x: float | ndarray, y: float | ndarray, z: float | ndarray
) -> tuple[float | ndarray, float | ndarray, float | ndarray]:
    """Transform Cartesian to spherical coordinates"""
    hxy = hypot(x, y)
    r = hypot(hxy, z)
    el = atan2(z, hxy)
    az = atan2(y, x)
    return az, el, r


def sph2cart(
    az: float | ndarray, el: float | ndarray, r: float | ndarray
) -> tuple[float | ndarray, float | ndarray, float | ndarray]:
    """Transform spherical to Cartesian coordinates"""
    rcos_theta = r * cos(el)
    x = rcos_theta * cos(az)
    y = rcos_theta * sin(az)
    z = r * sin(el)
    return x, y, z


def sanitize(lat: float | ndarray, ell: Ellipsoid, deg: bool) -> tuple[float | ndarray, Ellipsoid]:
    if ell is None:
        ell = Ellipsoid()
    if asarray is not None:
        lat = asarray(lat)
    if deg:
        lat = radians(lat)

    if asarray is not None:
        if (abs(lat) > pi / 2).any():
            raise ValueError("-pi/2 <= latitude <= pi/2")
    else:
        if abs(lat) > pi / 2:
            raise ValueError("-pi/2 <= latitude <= pi/2")

    return lat, ell



from pytest import approx
import pymap3d.vincenty as vincenty


def test_track2():
    lats, lons = vincenty.track2(40, 80, 65, -148, npts=3)
    assert lats == approx([40, 69.633139886, 65])
    assert lons == approx([80, 113.06849104, -148])



import pytest
from pytest import approx

import pymap3d as pm
import pymap3d.rcurve as rcurve

ell = pm.Ellipsoid()
A = ell.semimajor_axis


@pytest.mark.parametrize(
    "lat,curvature", [(0, A), (90, 0), (-90, 0), (45.0, 4517590.87884893), (-45, 4517590.87884893)]
)
def test_rcurve_parallel(lat, curvature):
    assert rcurve.parallel(lat) == approx(curvature, abs=1e-9)


def test_numpy_parallel():
    pytest.importorskip("numpy")
    assert rcurve.parallel([0, 90]) == approx([A, 0], abs=1e-9)


@pytest.mark.parametrize(
    "lat,curvature",
    [
        (0, 6335439.327),
        (90, 6399593.6258),
        (-90, 6399593.6258),
        (45.0, 6367381.8156),
        (-45, 6367381.8156),
    ],
)
def test_rcurve_meridian(lat, curvature):
    assert rcurve.meridian(lat) == approx(curvature)


def test_numpy_meridian():
    pytest.importorskip("numpy")
    assert rcurve.meridian([0, 90]) == approx([6335439.327, 6399593.6258])


def test_numpy_transverse():
    pytest.importorskip("numpy")
    assert rcurve.transverse([0, 90]) == approx([A, 6399593.626])



import pytest
from pytest import approx
from datetime import datetime

from pymap3d.timeconv import str2dt
import pymap3d.sidereal as pms

t0 = datetime(2014, 4, 6, 8)


def test_juliantime():
    assert pms.juliandate(t0) == approx(2.456753833333e6)


def test_types():
    np = pytest.importorskip("numpy")
    assert str2dt(t0) == t0  # passthrough
    assert str2dt("2014-04-06T08:00:00") == t0
    ti = [str2dt("2014-04-06T08:00:00"), str2dt("2014-04-06T08:01:02")]
    to = [t0, datetime(2014, 4, 6, 8, 1, 2)]
    assert ti == to  # even though ti is numpy array of datetime and to is list of datetime

    t1 = [t0, t0]
    assert (np.asarray(str2dt(t1)) == t0).all()


def test_datetime64():
    np = pytest.importorskip("numpy")
    t1 = np.datetime64(t0)
    assert str2dt(t1) == t0

    t1 = np.array([np.datetime64(t0), np.datetime64(t0)])
    assert (str2dt(t1) == t0).all()


def test_xarray_time():
    xarray = pytest.importorskip("xarray")

    t = {"time": t0}
    ds = xarray.Dataset(t)
    assert str2dt(ds["time"]) == t0

    t2 = {"time": [t0, t0]}
    ds = xarray.Dataset(t2)
    assert (str2dt(ds["time"]) == t0).all()


def test_pandas_time():
    pandas = pytest.importorskip("pandas")

    t = pandas.Series(t0)
    assert (str2dt(t) == t0).all()

    t = pandas.Series([t0, t0])
    assert (str2dt(t) == t0).all()



import pytest
from pytest import approx
from math import radians
from datetime import datetime

import pymap3d.sidereal as pmd
import pymap3d.haversine as pmh

lon = -148
t0 = datetime(2014, 4, 6, 8)
sra = 2.90658
ha = 45.482789587392013


@pytest.mark.parametrize(
    "time, use_astropy", [(t0, False), (t0, True), ([t0], False), ([t0], True)]
)
def test_sidereal(time, use_astropy):
    if use_astropy:
        pytest.importorskip("astropy")
    # http://www.jgiesen.de/astro/astroJS/siderealClock/
    tsr = pmd.datetime2sidereal(time, radians(lon), use_astropy=use_astropy)
    if isinstance(tsr, list):
        tsr = tsr[0]
    assert tsr == approx(sra, rel=1e-5)


def test_anglesep():
    pytest.importorskip("astropy")
    assert pmh.anglesep(35, 23, 84, 20) == approx(ha)


def test_anglesep_meeus():
    # %% compare with astropy
    assert pmh.anglesep_meeus(35, 23, 84, 20) == approx(ha)
